from pandas import DataFrame
from sqlalchemy import PoolProxiedConnection

from ..dataframe.analyzer import Analyzer
from .. import constants
import pandas as pd
import numpy as np
from ..logger import Logger
logger = Logger().get_logger()

class Validator:
    """
    Validates the upload of a DataFrame to a database table.

    Args:
        connection: The database connection object.
        df: The DataFrame to be uploaded.
        schema: The schema of the destination table.
        table: The name of the destination table.

    Raises:
        ExtraColumnsException: If the DataFrame has extra columns not present in the database table.
        ColumnDataException: If there are type mismatches or truncation issues with the columns in the DataFrame.
    """
    def __init__(self, connection: PoolProxiedConnection, df: pd.DataFrame, schema: str, table: str) -> None:
        self._connection = connection
        self._df = df
        self._schema = schema
        self._table = table

    @staticmethod
    def validate_upload(connection: PoolProxiedConnection, df: pd.DataFrame, schema: str, table: str)  -> None:
        df_metadata, column_info_df = Validator._fetch_column_info(connection, df, schema, table)
        Validator._check_extra_columns(df, column_info_df, schema, table)
        Validator._validate_column_types(df_metadata, column_info_df)

    @staticmethod
    def _fetch_column_info(connection: PoolProxiedConnection, df: pd.DataFrame, schema: str, table: str) -> tuple[
        list[dict], DataFrame]:
        get_column_info_query = (
            f'select COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH, NUMERIC_PRECISION '
            f'from INFORMATION_SCHEMA.columns '
            f'where table_schema = \'{schema}\' and table_name = \'{table}\'')
        column_info_df = pd.read_sql(get_column_info_query, connection)
        df_metadata = Analyzer.generate_column_metadata(df, None, None, 2)
        return df_metadata, column_info_df

    @staticmethod
    def _check_extra_columns(df, column_info_df, schema, table):
        db_columns = column_info_df['COLUMN_NAME'].tolist()
        new_columns = np.setdiff1d(df.columns.tolist(), db_columns)
        if new_columns.size > 0:
            extra_columns_df = df[new_columns]
            column_metadata = Analyzer.generate_column_metadata(extra_columns_df, None, None, 2)
            extra_columns_string = "\n".join([column.__str__() for column in column_metadata])
            type_mismatch_error_message = f'The table {schema}.{table} is missing the following columns:\n {extra_columns_string}'
            raise ExtraColumnsException(type_mismatch_error_message)

    @staticmethod
    def _validate_column_types(df_metadata, column_info_df):
        type_mismatch_columns = []
        truncated_columns = []

        for column in df_metadata:
            column_name = column['column_name']
            if column['is_empty']:
                logger.info(f'{column_name} is empty skipping type validation')
                continue
            db_column_info = column_info_df[column_info_df['COLUMN_NAME'] == column_name].iloc[0]
            db_column_data_type = db_column_info['DATA_TYPE']
            df_column_data_type = column['data_type']

            if Validator._is_type_mismatch(df_column_data_type, db_column_data_type):
                type_mismatch_columns.append(
                    f'{column_name} in dataframe is of type {df_column_data_type} while the database expects a type of {db_column_data_type}')
                continue

            if df_column_data_type in constants.DB_INT_TYPES + constants.DB_FLOAT_TYPES:
                truncate_message = Validator._check_numeric_truncation(column, db_column_info)
                if truncate_message is not None:
                    truncated_columns.append(truncate_message)
            elif df_column_data_type in constants.DB_DATE_TYPES + constants.DB_STR_TYPES:
                truncate_message = Validator._check_string_or_date_truncation(column, db_column_info)
                if truncate_message is not None:
                    truncated_columns.append(truncate_message)
        if type_mismatch_columns or truncated_columns:
            error_message = '\n'.join(type_mismatch_columns) + '\n'.join(truncated_columns)
            raise ColumnDataException(error_message)

    @staticmethod
    def _is_type_mismatch(df_column_data_type, db_column_data_type):
        for db_type in constants.DB_TYPES:
            if db_column_data_type in db_type and df_column_data_type not in db_type:
                return True
        return False

    @staticmethod
    def _check_numeric_truncation(column, db_column_info):
        df_numeric_precision = column['float_precision']
        db_column_numeric_precision = db_column_info['NUMERIC_PRECISION']
        if df_numeric_precision is None:
            return
        if df_numeric_precision > db_column_numeric_precision:
            return f'{column} needs a minimum of {df_numeric_precision} precision to be inserted'

    @staticmethod
    def _check_string_or_date_truncation(column, db_column_info):
        df_max_string_length = column['max_str_size']
        db_column_string_length = db_column_info.get('CHARACTER_MAXIMUM_LENGTH')
        if df_max_string_length is None:
            return
        if db_column_string_length == -1:
            return
        if df_max_string_length > db_column_string_length:
            return f'{column} needs a minimum of {df_max_string_length} size to be inserted'

    def validate(self):
        return self.validate_upload(self._connection, self._df, self._schema, self._table)

class ExtraColumnsException(Exception):
    """
    This class represents an exception that is raised when there are extra columns
    in a dataset that are not expected.

    :param Exception: The base exception class.
    """
    pass


class ColumnDataException(Exception):
    """
    Defines the ColumnDataException class, which is an exception subclass used for raising errors related to column data.

    Classes:
        ColumnDataException(Exception): An exception subclass for column data errors.
    """
    pass



import itertools
from typing import Hashable
import pandas as pd
import math
from .parser import Parser


class Analyzer:
    @staticmethod
    def find_unique_columns(df: pd.DataFrame) -> list[Hashable]:
        total_records = df.shape[0]
        unique_columns = []
        for column, series in df.items():
            column_unique = series.unique()
            column_unique_count = column_unique.size
            if column_unique_count == total_records:
                unique_columns.append(column)
        return unique_columns

    @staticmethod
    def find_unique_column_pairs(df: pd.DataFrame) -> list[tuple[Hashable, Hashable]]:
        total_records = df.shape[0]
        column_list = df.columns
        unique_columns = Analyzer.find_unique_columns(df)
        unique_column_pairs = []
        combo_df = pd.DataFrame()
        for column_set in itertools.combinations(column_list, 2):
            if column_set is None:
                continue
            first_column = column_set[0]
            second_column = column_set[1]
            if first_column in unique_columns or second_column in unique_columns:
                continue
            combo_df["combo"] = df[first_column].astype(str) + df[second_column].astype(str)
            combined_unique = combo_df["combo"].unique()
            combined_unique_count = combined_unique.size
            if combined_unique_count == total_records:
                unique_column_pairs.append(column_set)
        return unique_column_pairs

    @staticmethod
    def find_empty_columns(df: pd.DataFrame) -> list[str]:
        empty_columns = []
        for column, series in df.items():
            if series.dropna().empty:
                empty_columns.append(column.__str__())
        return empty_columns

    @staticmethod
    def generate_column_metadata(df: pd.DataFrame, primary_key: str, unique_columns: list[str], decimal_places: int) -> list[dict]:
        column_metadata_list = []
        if df.empty:
            return []
        for column, series in df.items():
            column_metadata = {
                'column_name': column,
                'data_type': None,
                'is_id': column == primary_key,
                'is_unique': unique_columns and column in unique_columns,
                'is_empty': False,
                'max_str_size': None,
                'float_precision': None,
                'decimal_places': None,
                'biggest_num': None,
                'smallest_num': None
            }
            if series.dropna().empty:
                column_metadata['is_empty'] = True
                column_metadata_list.append(column_metadata)
                continue
            try:
                series.apply(Parser.parse_float)
                no_null_series = series.dropna()
                if not no_null_series.eq(0).all():
                    left_digits = int(math.log10(abs(series.max()))) + 1
                    float_precision = left_digits + decimal_places
                    column_metadata['data_type'] = 'float'
                    column_metadata['float_precision'] = float_precision
                    column_metadata['decimal_places'] = decimal_places
                series.apply(Parser.parse_integer)
                biggest_num = series.max()
                smallest_num = series.min()
                column_metadata['data_type'] = 'integer'
                column_metadata['biggest_num'] = biggest_num
                column_metadata['smallest_num'] = smallest_num
                column_metadata['float_precision'] -= decimal_places
            except (ValueError, TypeError):
                pass
            try:
                series.apply(Parser.parse_boolean)
                column_metadata['data_type'] = 'boolean'
            except ValueError:
                pass
            if column_metadata['data_type'] is None:
                try:
                    series.apply(Parser.parse_date)
                    column_metadata['data_type'] = 'datetime'
                except (ValueError, TypeError, OverflowError):
                    pass
            if column_metadata['data_type'] is None:
                str_series = series.apply(str)
                largest_string_size = str_series.str.len().max()
                column_metadata['data_type'] = 'string'
                column_metadata['max_str_size'] = largest_string_size
            column_metadata_list.append(column_metadata)
        return column_metadata_list

    @staticmethod
    def find_categorical_columns(df: pd.DataFrame, unique_threshold: float = 1) -> list[Hashable]:
        if unique_threshold < 0 or unique_threshold > 1:
            raise ValueError('Unique threshold must be between 0 and 1')
        categorical_columns = []
        for column, series in df.items():
            no_null_series = series.dropna()
            if no_null_series.empty:
                continue
            column_count = no_null_series.size
            column_unique_count = no_null_series.unique().size
            unique_pct = column_unique_count / column_count
            # print(f'{column} has %{unique_pct*100} unique values')
            if unique_pct <= unique_threshold:
                categorical_columns.append(column)
        return categorical_columns



import unittest
import pandas as pd
from datetime import datetime
from src.etl.query.creator import Creator


class TestTableMaker(unittest.TestCase):

    def test_make_mssql_table(self):
        df = pd.DataFrame({
            'id_column': [2147483649, 2147483650, 2147483651],
            'int_column': [1, 2, 3],
            'float_column': [1.1, 2.2, 3.3],
            'date_column': ['2021-01-01', '2021-01-02', '2021-01-03'],
            'bool_column': [True, False, True],
            'str_column': ['a', 'bb', 'ccc'],
            'empty_column': [None, None, None]
        })

        expected_query = (
            "create table dbo.[test_table]\n(\n"
            "[id_column] bigint constraint pk_test_table_id_column primary key,\n"
            "\t[int_column] tinyint,\n"
            "\t[float_column] decimal(10, 2),\n"
            "\t[date_column] datetime2,\n"
            "\t[bool_column] bit,\n"
            "\t[str_column] nvarchar(23),\n"
            "\t[empty_column] nvarchar(max),\n"
            "\tsystem_record_start datetime2 generated always as row start\n"
            "\t\tconstraint df_test_table_system_record_start\n"
            "\t\tdefault sysutcdatetime() not null,\n"
            "\tsystem_record_end datetime2 generated always as row end\n"
            "\t\tconstraint df_test_table_system_record_end\n"
            "\t\tdefault sysutcdataetime() not null,\n"
            "\t\tperiod for system_time(system_record_start, system_record_end)\n)"
            " with \n(\tsystem_versioning = on (history_table = dbo.[test_table_history])\n);"
        )

        actual_query = Creator.create_mssql_table(df, 'dbo', 'test_table',
                                                  primary_key='id_column', history=True)

        self.assertEqual(expected_query, actual_query)


if __name__ == '__main__':
    unittest.main()



# test_loader.py
import unittest
from unittest.mock import Mock, patch

import pandas as pd
from src.etl.database.loader import Loader
from sqlalchemy.engine.interfaces import DBAPICursor


class TestLoader(unittest.TestCase):

    def setUp(self):
        self.cursor = Mock(spec=DBAPICursor)
        self.df = pd.DataFrame({'col1': [1, 2], 'col2': [3, 4]})
        self.schema = 'test_schema'
        self.table = 'test_table'
        self.loader = Loader(self.cursor, self.df, self.schema, self.table)
        self.column_string = 'col1, col2'
        self.placeholders = ['%s', '%s']
        self.location = f"{self.schema}.{self.table}"

    @patch('src.etl.database.loader.insert_to_db')
    def test_insert_to_table_with_small_dataframe(self, mock_insert_to_db):
        Loader._insert_to_table(self.column_string, self.cursor, self.df, self.location, self.placeholders)
        mock_insert_to_db.assert_called_once()
        self.assertEqual(mock_insert_to_db.call_args[0][0], self.column_string)
        self.assertEqual(mock_insert_to_db.call_args[0][1], self.cursor)

    @patch('src.etl.database.loader.insert_to_db')
    def test_insert_to_table_with_empty_dataframe(self, mock_insert_to_db):
        empty_df = pd.DataFrame(columns=['col1', 'col2'])
        Loader._insert_to_table(self.column_string, self.cursor, empty_df, self.location, self.placeholders)
        mock_insert_to_db.assert_not_called()

    @patch('src.etl.database.loader.insert_to_db')
    def test_insert_to_table_raises_exception(self, mock_insert_to_db):
        error_message = "test error"
        mock_insert_to_db.side_effect = Exception(error_message)
        with self.assertRaises(Exception) as context:
            Loader._insert_to_table(self.column_string, self.cursor, self.df, self.location, self.placeholders)
        self.assertTrue(error_message in str(context.exception))


if __name__ == '__main__':
    unittest.main()



import pandas as pd
from pandas import DataFrame
from sqlalchemy import PoolProxiedConnection


class DatabaseUtils:
    """
    Utility class for database operations, such as fetching table and column data.
    """

    def __init__(self, connection: PoolProxiedConnection):
        self.connection = connection

    def get_table_list(self, schema: str) -> list:
        query = (
            f"SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES "
            f"WHERE TABLE_SCHEMA = '{schema}' AND TABLE_TYPE = 'BASE TABLE';"
        )
        return pd.read_sql(query, self.connection)['TABLE_NAME'].tolist()

    def get_column_names_and_types(self, schema: str, table: str) -> DataFrame:
        query = (
            f"SELECT COLUMN_NAME, DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS "
            f"WHERE TABLE_SCHEMA = '{schema}' AND TABLE_NAME = '{table}';"
        )
        return pd.read_sql(query, self.connection)

    def get_column_data(self, schema: str, table: str, column: str) -> pd.Series:
        query = f"SELECT DISTINCT [{column}] FROM {schema}.{table}"
        return pd.read_sql(query, self.connection)[column].dropna()



import pandas as pd
from ..dataframe.analyzer import Analyzer
from ..logger import Logger
logger = Logger().get_logger()


class Creator:
    def __init__(self, df: pd.DataFrame, schema: str, table: str, primary_key: str = None,
                 unique_columns: list[str] = None, history: bool = False,
                 varchar_padding: int = 20, float_precision: int = 10, decimal_places: int = 2,
                 generate_id: bool = False) -> None:
        self._df = df
        self._schema = schema
        self._table = table
        self._primary_key = primary_key
        self._unique_columns = unique_columns
        self._history = history
        self._varchar_padding = varchar_padding
        self._float_precision = float_precision
        self._decimal_places = decimal_places
        self._generate_id = generate_id

    @staticmethod
    def create_mssql_table(df: pd.DataFrame, schema: str, table: str, primary_key: str = None,
                           unique_columns: list[str] = None, history: bool = False,
                           varchar_padding: int = 20, float_precision: int = 10, decimal_places: int = 2,
                           generate_id: bool = False) -> str:
        location = f'{schema}.[{table}]'
        column_metadata = Analyzer.generate_column_metadata(df, primary_key, unique_columns, decimal_places)
        if len(column_metadata) < 1:
            return ''
        column_type_list = []
        if generate_id:
            id_string = f'id int identity constraint pk_{table}_id primary key'
            column_type_list.append(id_string)
        for column in column_metadata:
            column_string = None
            column_name = column['column_name']
            if column['is_empty']:
                logger.info(f"{column_name} is empty - setting to nvarchar(max)")
                column_string = f'[{column_name}] nvarchar(max)'
                column_type_list.append(column_string)
                continue
            if column['data_type'] == 'datetime':
                column_string = f'[{column_name}] datetime2'
            if column['data_type'] == 'float':
                if float_precision < column['float_precision']:
                    float_precision = column['float_precision']
                column_string = f'[{column_name}] decimal({float_precision}, {decimal_places})'
            if column['data_type'] == 'integer':
                if column['smallest_num'] < -2147483648 or column['biggest_num'] > 2147483648:
                    column_string = f'[{column_name}] bigint'
                if column['smallest_num'] >= -2147483648 and column['biggest_num'] <= 2147483648:
                    column_string = f'[{column_name}] int'
                if column['smallest_num'] >= -32768 and column['biggest_num'] <= 32768:
                    column_string = f'[{column_name}] smallint'
                if column['smallest_num'] >= 0 and column['biggest_num'] <= 255:
                    column_string = f'[{column_name}] tinyint'
            if column['data_type'] == 'boolean':
                column_string = f'[{column_name}] bit'
            if column['data_type'] == 'string':
                padded_length = int(column['max_str_size'] + varchar_padding)
                if padded_length >= 4000:
                    column_string = f'[{column_name}] nvarchar(max)'
                else:
                    column_string = f'[{column_name}] nvarchar({padded_length})'
            if column['is_id']:
                column_string += f' constraint pk_{table}_{column_name} primary key'
            if column['is_unique']:
                column_string += f' constraint ak_{table}_{column_name} unique'
            column_type_list.append(column_string)

        column_type_string = ',\n\t'.join(column_type_list)
        create_query = f'create table {location}\n(\n\t{column_type_string}\n);'
        if history:
            history_location = f'{schema}.[{table}_history]'
            history_insert = (
                f'{column_type_string},\n'
                '\tsystem_record_start datetime2 generated always as row start\n'
                f'\t\tconstraint df_{table}_system_record_start\n'
                '\t\tdefault sysutcdatetime() not null,\n'
                '\tsystem_record_end datetime2 generated always as row end\n'
                f'\t\tconstraint df_{table}_system_record_end\n'
                '\t\tdefault sysutcdataetime() not null,\n'
                '\t\tperiod for system_time(system_record_start, system_record_end)'
            )
            create_query = (
                f'create table {location}\n(\n{history_insert}\n) with \n('
                f'\tsystem_versioning = on (history_table = {history_location})\n);'
            )
        return create_query

    @staticmethod
    def create_mariadb_table(df: pd.DataFrame, schema: str, table: str, primary_key: str = None,
                             unique_columns: list[str] = None, history: bool = False,
                             varchar_padding: int = 20, float_precision: int = 10, decimal_places: int = 2,
                             generate_id: bool = False) -> str:
        location = f'{schema}.{table}'
        column_metadata = Analyzer.generate_column_metadata(df, primary_key, unique_columns, decimal_places)
        column_type_list = []
        if generate_id:
            id_string = f'id int auto_increment, constraint pk_{table}_id primary key (id)'
            column_type_list.append(id_string)
        for column in column_metadata:
            column_string = None
            column_name = column['column_name']
            if column['is_empty']:
                logger.info(f"{column_name} is empty - skipping")
                continue
            if column['data_type'] == 'datetime':
                column_string = f'`{column_name}` datetime'
            if column['data_type'] == 'float':
                if float_precision < column['float_precision']:
                    float_precision = column['float_precision']
                column_string = f'`{column_name}` decimal({float_precision}, {decimal_places})'
            if column['data_type'] == 'integer':
                if column['smallest_num'] < -2147483648 or column['biggest_num'] > 2147483648:
                    column_string = f'`{column_name}` bigint'
                if column['smallest_num'] >= -2147483648 and column['biggest_num'] <= 2147483648:
                    column_string = f'`{column_name}` int'
                if column['smallest_num'] >= -32768 and column['biggest_num'] <= 32768:
                    column_string = f'`{column_name}` smallint'
                if column['smallest_num'] >= -128 and column['biggest_num'] <= 127:
                    column_string = f'`{column_name}` tinyint'
            if column['data_type'] == 'boolean':
                column_string = f'`{column_name}` bit'
            if column['data_type'] == 'string':
                padded_length = int(column['max_str_size'] + varchar_padding)
                if padded_length > 21844:
                    logger.info(f"{column_name} has data too large for storing - skipping")
                    continue
                else:
                    column_string = f'`{column_name}` varchar({padded_length})'
            if column['is_id']:
                key_string = f'constraint pk_{table}_{column_name} primary key ({column_name})'
                column_type_list.append(key_string)
            if column['is_unique']:
                key_string = f'constraint ak_{table}_{column_name} unique ({column_name})'
                column_type_list.append(key_string)
            column_type_list.append(column_string)

        column_type_string = ', '.join(column_type_list)
        create_query = f'create table {location} ({column_type_string});'
        if history:
            create_query = (
                f'create table {location} ({column_type_string}) with ('
                f'system_versioning;'
            )
        return create_query

    # generate convenience methods that use the classes variables and above static methods
    def new_mssql_table(self) -> str:
        return Creator.create_mssql_table(self._df, self._schema, self._table, self._primary_key, self._unique_columns,
                                          self._history, self._varchar_padding, self._float_precision,
                                          self._decimal_places, self._generate_id)

    def new_mariadb_table(self) -> str:
        return Creator.create_mariadb_table(self._df, self._schema, self._table, self._primary_key,
                                            self._unique_columns, self._history, self._varchar_padding,
                                            self._float_precision, self._decimal_places, self._generate_id)



class MsSqlUpdater:

    def __init__(self, source_schema: str, source_table: str, source_columns: list[str], source_id_column: str,
                 target_schema: str, target_table: str, target_columns: list[str], target_id_column: str):
        self._source_schema = source_schema
        self._source_table = source_table
        self._source_columns = source_columns
        self._source_id_column = source_id_column
        self._target_schema = target_schema
        self._target_table = target_table
        self._target_columns = target_columns
        self._target_id_column = target_id_column

    @staticmethod
    def merge_sql(source_schema: str, source_table: str, source_columns: list[str], source_id_column: str,
                  target_schema: str, target_table: str, target_columns: list[str], target_id_column: str,
                  delete_unmatched: bool = True) -> str:
        if len(source_columns) != len(target_columns):
            raise ValueError("source_columns and target_columns must have the same length")
        stage = f'{source_schema}.{source_table}'
        target_id_column = f'[{target_id_column}]'
        source_id_column = f'[{source_id_column}]'

        location = f'{target_schema}.{target_table}'
        clean_target_columns = [f'[{column}]' for column in target_columns]
        clean_source_columns = [f'[{column}]' for column in source_columns]

        target_columns_str = ', '.join([f'{column}' for column in clean_target_columns])
        source_columns_str = ', '.join([f'b.{column}' for column in clean_source_columns])
        comparison_list = [(src_col, tgt_col) for src_col, tgt_col in zip(clean_source_columns, clean_target_columns)]
        comparison_str = ' or '.join(
            [f'a.{column[0]} <> b.{column[1]} or (a.{column[0]} is null and b.{column[1]} is not null) ' for column in
             comparison_list if column[0] != target_id_column]
        )
        update_str = ',\n\t\t'.join(
            [f'a.{column[0]} = b.{column[1]}' for column in comparison_list if column[0] != target_id_column])
        query = (
            f'merge {location} a\n'
            f'using {stage} b\n'
            f'on a.{target_id_column} = b.{source_id_column}\n'
            f'when matched and ({comparison_str}) then\n'
            f'\tupdate\n'
            f'\tset {update_str}\n'
            f'when not matched by target then\n'
            f'\tinsert ({target_columns_str})\n'
            f'\tvalues ({source_columns_str})'
        )
        if delete_unmatched:
            query = f'{query}\nwhen not matched by source then delete'
        return f'{query};'

    @staticmethod
    def upsert_sql(source_schema: str, source_table: str, source_columns: list[str], source_id_column: str,
                   target_schema: str, target_table: str, target_columns: list[str], target_id_column: str) -> str:
        stage = f'{source_schema}.{source_table}'
        location = f'{target_schema}.{target_table}'
        clean_target_columns = [f'[{column}]' for column in target_columns]
        clean_source_columns = [f'[{column}]' for column in source_columns]
        target_column_string = ', '.join(clean_target_columns)
        source_column_string = ', '.join(clean_source_columns)

        stage_columns = [f's.{column}' for column in clean_source_columns]
        stage_column_string = ', '.join(stage_columns)
        delete_dupes_query = (
            f'Delete from {stage} from {stage} s where exists (select '
            f'{stage_column_string} intersect select {target_column_string} from {location})'
        )
        delete_old_query = (
            f'delete from {location} where {target_id_column} in ( '
            f'select {source_id_column} from {stage} intersect select {target_id_column} from {location})'
        )
        insert_query = (
            f'insert into {location} ({target_column_string}) select {source_column_string} from {stage}'
        )
        query = f'{delete_dupes_query}; {delete_old_query}; {insert_query};'
        return query

    @staticmethod
    def append_sql(source_schema: str, source_table: str, source_columns: list[str], target_schema: str,
                   target_table: str, target_columns: list[str]) -> str:
        stage = f'{source_schema}.{source_table}'
        location = f'{target_schema}.{target_table}'
        clean_target_columns = [f'[{column}]' for column in target_columns]
        clean_source_columns = [f'[{column}]' for column in source_columns]

        target_column_string = ','.join(clean_target_columns)
        source_column_string = ','.join(clean_source_columns)

        query = (
            f'insert into {location} ({target_column_string}) select {source_column_string} from {stage}'
            f' except select {target_column_string} from {location}'
        )
        return query

    def merge(self, delete_unmatched: bool = True) -> str:
        return self.merge_sql(
            self._source_schema, self._source_table, self._source_columns, self._source_id_column,
            self._target_schema, self._target_table, self._target_columns, self._target_id_column, delete_unmatched
        )

    def upsert(self) -> str:
        return self.upsert_sql(
            self._source_schema, self._source_table, self._source_columns, self._source_id_column,
            self._target_schema, self._target_table, self._target_columns, self._target_id_column
        )

    def append(self) -> str:
        return self.append_sql(
            self._source_schema, self._source_table, self._source_columns,
            self._target_schema, self._target_table, self._target_columns
        )